#!/usr/bin/python -u
## Change the above to match your site
## The option -u is for unbuffered output

## Import required libraries
from malib import *
import sys
import os

## To run this file, use the command line options:
## rotorA.py term alus mults iterates mix
## where term is the string term and dumps output to the terminal
## where alus is number of alus (an integer)
## where mults is the number of multipliers (an integer)
## where iterates is the number of iterates (an integer) (iterates pg. 63- of ABSS dissertation)
## where mix is the name of a python file describing what control paths to optimize
##
## A good initial trial is
## rotorA.py term 1 2 1 mixall.py


## This particular version of rotor originated with Ivan
## the file rotor.ps describes the algorithm (negate requires a resource use and time)

outname = sys.argv[1]

if not outname == "term":
        print "Output to",outname
        sys.stdout = open(outname,"w")

ialu = int(sys.argv[2])
imult = int(sys.argv[3])
iiterates = int(sys.argv[4])
mix = sys.argv[5]

print os.times()

## A CMA (refer to my dissertation) is built for each benchmark
## This corresponds to a python class derived from the base model class in ma.py
## This is an iterate, another class will compose copies of this specific iterate
class Rotor(Model):

    ## The constructs the model
    def __init__(self,parent):

        ## Must manually call the base constructor
        Model.__init__(self,parent)

        ## This is the list of all submodels (MA) in this CMA
        self.submodels = []

        # Modeling NFA, BDD constructed as ordered here

        ## This is the sense MA required for cyclic CDFGs as described on pg. 66, ABSS dissertation
        self.sense = COp1( self )


        ## self.l5 is a model of type COp1 where COp1 is defined in malib.py
        ## COp1 one is a resource with single time step sequence behavior
        self.l5 = COp1( self ) 
        self.s4 = COp1( self ) 
        self.s3 = COp1( self )
        self.s2 = COp1( self ) 
        self.l2 = COp1( self )

        self.c1  = COp1( self )
        ## We need to know the value of the operand produced by self.c1 (control)
        ## self.c1c is the multivalue MA synchronized with self.c1. Described pg. 36-41 ABSS dissertation
        self.c1c = Case( self, 2 )
        
        self.l1 = COp1( self ) 
        self.l3 = COp1( self ) 
        self.c2 = COp1( self )
        self.c2c = Case( self, 2 )
        self.s1 = COp1( self ) 
        self.s6 = COp1( self ) 
        self.l6 = COp1( self ) 
        self.s7 = COp1( self ) 
        self.s8 = COp1( self ) 
        self.l7 = COp1( self ) 
        self.c3 = COp1( self )
        self.c3c = Case( self, 2 )
        self.l9 = COp1( self ) 
        self.s11 = COp1( self ) 
        self.s5 = COp1( self )    
        self.s10 = COp1( self ) 
        self.l8 = COp1( self ) 
        self.s9 = COp1( self )   
        self.m1 = COp2p( self )
        self.m2 = COp2p( self )
        self.a1 = COp1( self )   
        self.m3 = COp2p( self ) 
        self.m4 = COp2p( self ) 
        self.a2 = COp1( self ) 

        ## For debugging -- This reverse maps the names to the MA to make printouts pretty
        ## See the code in ma.py
        self.AddNames()

    ## CDFGs are described differently than DFGs such as ewf-1.py
    ## Where DFGs use an edge-centric spec, CDFGs use a task-centric spec
    ## The CDFG spec is what is described in the dissertation
    def BuildOperandLinks(self):

        ## The sense operand MA has no inputs, hence one empty tupe
        self.sense.oinputs = ( (), )

        ## The task s1 has no defined inputs
        self.s1.oinputs = ( (), )

        ## DT(self.s1, 0, 0) returns an dependency tuple with source operand from self.s1 index 0
        ## The last 0 indicates sense, in this case not inverted or intraiteration
        ## The accepter of this operan is task c1,
        ## This is 2-deep nested tuple
        ## The outer-most tuple items are indexed to inputs of the accepter
        ## In this case, there is only one input operand tuple at position 0 and hence maps
        ## to task self.c1 input 0
        ## The inner-most tuples list all possible sources for this operand. In this case there is no
        ## join so only one source, self.s1, is used to create a dependency tuple
        self.c1.oinputs = ( ( DT(self.s1, 0, 0), ), )
        self.s2.oinputs = ( (), )
        self.c2.oinputs = ( ( DT(self.s2, 0, 0), ), )
        self.l1.oinputs = ( (), )
        self.l2.oinputs = ( ( DT(self.s2, 0, 0), ), )
        self.l3.oinputs = ( ( DT(self.s1, 0, 0), ), )
        self.s3.oinputs = ( ( DT(self.s2, 0, 0), ), )
        self.l5.oinputs = ( ( DT(self.s3, 0, 0), ), )
        self.s4.oinputs = ( ( DT(self.l5, 0, 0), ), )
        self.s5.oinputs = ( (), )
        self.c3.oinputs = ( ( DT(self.s5, 0, 0), ), )
        self.s6.oinputs = ( ( DT(self.s1, 0, 0), ), )
        self.l6.oinputs = ( ( DT(self.s6, 0, 0), ), )
        self.s7.oinputs = ( ( DT(self.l6, 0, 0), ), )
        self.l7.oinputs = ( ( DT(self.s5, 0, 0), ), )
        self.s8.oinputs = ( ( DT(self.l7, 0, 0), ), )
        self.s9.oinputs = ( (), )
        self.l8.oinputs = ( ( DT(self.s9, 0, 0), ), )
        self.s10.oinputs = ( ( DT(self.l8, 0, 0), ), )
        self.s11.oinputs = ( ( DT(self.s5, 0, 0), ), )
        self.l9.oinputs = ( ( DT(self.s11, 0, 0), ), )

        ## Here is an example of tasks with operand resolution points
        ## The innermost tuple contains four dependency tuples, one of which will be selected
        ## to satisfy this input dependency
        ## Only dependencies for one input to this multiplier are specified. The other is left undetermined as it is a constant input
        self.m1.oinputs = ( ( DT(self.l2, 0, 0), DT(self.s4, 0, 0), DT(self.s8, 0, 0), DT(self.l9, 0, 0) ), )
        self.m2.oinputs = ( ( DT(self.l1, 0, 0), DT(self.l3, 0, 0), DT(self.s7, 0, 0), DT(self.s10, 0, 0) ), )

        self.m4.oinputs = ( ( DT(self.l2, 0, 0), DT(self.s4, 0, 0), DT(self.s8, 0, 0), DT(self.l9, 0, 0) ), )
        self.m3.oinputs = ( ( DT(self.l1, 0, 0), DT(self.l3, 0, 0), DT(self.s7, 0, 0), DT(self.s10, 0, 0) ), )

        ## self.a1 is a task with dependency tuples specified for two inputs: indices 0 and 1
        ## There are no operand resolution points for self.a1
        self.a1.oinputs = ( ( DT(self.m1, 0, 0), ),
                            ( DT(self.m2, 0, 0), ) )

        self.a2.oinputs = ( ( DT(self.m3, 0, 0), ),
                            ( DT(self.m4, 0, 0), ) )

        ## Control specific
        ## These are multivalue MA
        ## Output operands from these MA are not used in dependency tuples but rather to control obviate other tasks
        ## For this reason, the ooutputs (observable outputs) must be specified here and specified as empty
        ## These multivalue MA require no input operands, they are syncronized to their data-path mate MA who may
        ## accept and produce regular operands.
        self.c1c.ooutputs = ( (), )
        self.c1c.oinputs = ( (), )

        self.c2c.ooutputs = ( (), )
        self.c2c.oinputs = ( (), )       

        self.c3c.ooutputs = ( (), )
        self.c3c.oinputs = ( (), )


    ## This method builds the required scheduling problem control structure
    def BuildControlStructure(self):

        ## Define forks
        ## A list of all multivalue MA in the scheduling problem
        self.forks = [self.c1c, self.c2c, self.c3c]

        ## Initialize an empty set
        set = []
		
        # Synchronize multivalue control automata with mated datapath tasks
        ## i.e. c1c is synchronized with the production of operand 0 from self.c1
        self.c1c.Sync(set, self.c1, 0)
        self.c2c.Sync(set, self.c2, 0)
        self.c3c.Sync(set, self.c3, 0)

        ## Create a ROBDD vector from this set
        vec = List2Vec(set)
        ## And all terms in the vector
        syncterm = vec.And()

        ## Prune the relation in the 0 partition of transition relation structure through intersection
        self.TR[0].rel = self.TR[0].rel & syncterm

        ## Assign tasks to control blocks
        ## specify ncb for global operations
        ## Some tasks do not belong in any particular control block -- i.e. they are always executed
        ## These require a don't care control block which is specified here manually
        ## State(-2) instantiates an empty state.
        ## ncb is not in control block and these global tasks are in every control block (not empty)
        ## The list contains two items for both iteration senses
        ncb = [State(-2), State(-2)]
        self.sense.ncb = ncb
        self.s1.ncb = ncb
        self.c1.ncb = ncb
        self.c1c.ncb = ncb
        self.m1.ncb = ncb
        self.m2.ncb = ncb
        self.m3.ncb = ncb
        self.m4.ncb = ncb
        self.a1.ncb = ncb
        self.a2.ncb = ncb

        ## This identifies tasks which only occur under certain control outcomes
        ## The following may be read as, the tasks self.s2, self.c2....self.s4 are only necessary
        ## before and when the value of c1c is resolved to 0
        self.c1c.OnlyWhen( (0, ), (self.s2, self.c2, self.c2c, self.l1, self.l2, self.l3, self.s3, self.l5, self.s4) )

        ## The tasks self.s5, self.c3, ...., self.l9 are only necessary before and after the value of c1c is resolved to 1
        self.c1c.OnlyWhen( (1, ), (self.s5, self.c3, self.c3c, self.s6, self.l6, self.s7, self.s8, self.l7, self.s9, self.l8, self.s10, self.s11, self.l9) )

        ## Notice that tasks self.l1 and self.l2 which appeared two lines above also appear in this control block grouping for self.c2c.
        self.c2c.OnlyWhen1( (0, ), (self.l1, self.l2) )
        self.c2c.OnlyWhen1( (1, ), (self.l3, self.s3, self.l5, self.s4) )

        self.c3c.OnlyWhen1( (0, ), (self.s6, self.l6, self.s7, self.l7, self.s8) )
        self.c3c.OnlyWhen1( (1, ), (self.s9, self.l8, self.s10, self.s11, self.l9) )

    ## This method constructs the two-phase transiiton relation: TR=data phase cTR=control phase
    def BuildTRcTR(self):

        self.BuildControlStructure()
        self.BuildOperandLinks()

        ## Fill in the output operand information for each task
        for sm in self.submodels:
            sm.BuildOOutputs()

        ## Identify taskstart and taskfinal states
        ss0 = self.taskstart[0][0]
        ss1 = self.taskstart[0][1]

        ## Smooth out the sense operand MA
        ss0.Smooth(self.sense.smth)
        ss1.Smooth(self.sense.smth)

        ## Add the sense synchronization as described on pg 67, ABSS dissertation
        self.TR[0].rel = self.TR[0].rel & ( self.sense.required[0][0] >> (ss0.P | ss0.N) )
        self.TR[0].rel = self.TR[0].rel & ( self.sense.required[0][1] >> (ss1.P | ss1.N) )

        ## create empty lists
        setTR  = []
        setcTR = []

        ## Insist on one sense (allowed odd/even iterations depenend on instruction sense)
        ## additional sense constraints for each MA as on pg 67, ABSS dissertation
        for sm in self.submodels:
            if not (sm == self.sense):
                setTR.append( (sm.taskstart[0][0].P & ~sm.taskstart[0][0].N) >> self.sense.known[0][0].P )
                setTR.append( (sm.taskstart[0][1].P & ~sm.taskstart[0][1].N) >> self.sense.known[0][1].P )
            else:
                print "Found sense"

        ## Build dependency and capacity constraints for each MA in the composition
        for sm in self.submodels:
            sm.BuildTRcTRConstraints(setTR, setcTR)

        ## Prune the TR relation
        vec = List2Vec(setTR)
        self.TR[0].rel = self.TR[0].rel & vec.And()

        ## Prune the cTR relation
        vec = List2Vec(setcTR)
        self.cTR[0].rel = self.cTR[0].rel & vec.And()

        ## Top level hold relation
        self.hold = self.TR.vecps.HoldTR(self.TR.vecns)


    ## This method builds a complete iterate
    def BuildComplete(self,base):

        print "Complete Build"
        self.BuildCompositeState(base)
        Model.BuildCompositeTR(self)
        Model.BuildCompositecTR(self)
        
        print "Initial CMA TR and cTR:"
        self.TR.Info()
        self.cTR.Info()

        self.TR.SetVarMap()
        self.BuildVectors((0,0),(0,1)) # Don't flip guards
        self.BuildTRcTR()
        print "Dependency Refined CMA TR and cTR:"
        self.TR.Info()
        self.cTR.Info()

        self.DefineGroups()

        ## Do the viablity prune
        rs = self.DualReachTRcTR(self.taskstart[0])
        self.TR.PruneS(rs)
        ##self.cTR.PruneS(rs)
        print "Reachable Refined CMA TR and cTR:"     
        self.TR.Info()
        self.cTR.Info()

    ## This method builds a copy of an iterate that is shifted to a new range of ROBDD indices
    def BuildShifted(self,base,original):

        print "Shifted Build"
        self.BuildCompositeState(base)

        ## Build copies of the original transition relation objects
        self.TR = Trans(0,original.TR)
        self.cTR = Trans(0,original.cTR)

        ## Build a BDD shift object, this can handle ranges that overlap
        shft = ShiftBDD(original.base,original.width+original.base,base)

        ## Shift the TR objects
        self.TR.Shift(shft)
        self.cTR.Shift(shft)

        ## Create a new shifted taskstart list
        self.taskstart = [[],]
        for ss in original.taskstart[0]:
	    ## Copy state
            temp = State(-1,ss)
            temp.Shift(shft)
            self.taskstart[0].append(temp)

        ## Shift memory care conditions
        i = 0
        self.dual = mgr.One()
        for sm in self.submodels:

            ## Shift dual
            sm.dual = shft.Shift(original.submodels[i].dual)
            self.dual = self.dual & sm.dual
            try:
                sm.memcare[0][0] = shft.Shift(original.submodels[i].memcare[0][0])
                sm.memcare[0][1] = shft.Shift(original.submodels[i].memcare[0][1])
            except AttributeError:
                print "No memorycare for",sm.name
            i = i + 1
        
        ## Define Forks
        self.forks = [self.c1c, self.c2c, self.c3c]
        self.DefineGroups()

    ## This method groups busy labels according to resource classes
    def DefineGroups(self):

        ## This is the group of all activity that requires an alu
        self.alu_grp = [self.s1.busy[0], self.s2.busy[0], self.s3.busy[0], self.s4.busy[0], self.s5.busy[0], self.s6.busy[0], self.s7.busy[0], self.s8.busy[0], self.s9.busy[0], self.s10.busy[0], self.s11.busy[0], self.a1.busy[0], self.a2.busy[0] ]

        self.cond_grp = [ self.c1.busy[0], self.c2.busy[0], self.c3.busy[0] ]
        
        self.table_grp = [ self.l1.busy[0], self.l2.busy[0], self.l3.busy[0], self.l5.busy[0], self.l6.busy[0], self.l7.busy[0], self.l8.busy[0], self.l9.busy[0] ]

        self.mult_grp = [ self.m1.busy[0], self.m2.busy[0], self.m3.busy[0], self.m4.busy[0] ]

        ## One pool of registers in this example
        self.mem_grp = [ self.s1.memcare[0], self.s2.memcare[0], self.s3.memcare[0], self.s4.memcare[0], self.s5.memcare[0], self.s6.memcare[0], self.s7.memcare[0], self.s8.memcare[0], self.s9.memcare[0], self.s10.memcare[0], self.s11.memcare[0], self.l1.memcare[0], self.l2.memcare[0], self.l3.memcare[0], self.l5.memcare[0], self.l6.memcare[0], self.l7.memcare[0], self.l8.memcare[0], self.l9.memcare[0], self.m1.memcare[0], self.m2.memcare[0], self.m3.memcare[0], self.m4.memcare[0] ]

## This class builds a CMA that composes several Rotor iterates
class Top(Model):

    ## Constructor specifies number of iterates
    def __init__(self,parent,iterates):
		
        Model.__init__(self,parent)

        ## Submodels in this case are all the iterates in this top level CMA
        self.submodels = []

        ## The first iterate must be built completely
        self.submodels.append(Rotor(0))
        self.submodels[0].BuildComplete(0)

        ## The remaining iterates may be shifted copies of the first
        for i in range(1,iterates):
            self.submodels.append(Rotor(0))
            base = self.submodels[i-1].width+self.submodels[i-1].base
            self.submodels[i].BuildShifted(base,self.submodels[0])

        ## Define the base and width for this CMA
        self.base = 0
        self.width = self.submodels[iterates-1].base+self.submodels[iterates-1].width

    ## Build the TR for this top level CMA
    def BuildTR(self):

        ## Forks is a list of all forks from submodels
        self.forks = []
        for sm in self.submodels:
            self.forks = self.forks + sm.forks

	## Build resolve labeled state
        ## Identify the resolve state for the composition MA
        vec = DdArray(len(self.forks))
        self.resolve = State(-2)
        for frk in self.forks:
            self.resolve = self.resolve | frk.resolve

        ## Construct skeleton transition relations for this CMA
        ## Refer to ma.py for more details
        self.TR = Trans(1) 
        self.TR.PreCompose()

        ## An empty taskstart structure
        self.taskstart = ([State(0),State(0)], )

        ## Build the top-level CMA's from all member task start 
        for sm in self.submodels:
            self.TR.Add(sm.TR)
            for i in range(0,len(sm.taskstart[0])):
                try:
                    self.taskstart[0][i] = self.taskstart[0][i] & sm.taskstart[0][i]
                except IndexError:
                    self.taskstart[0].append(sm.taskstart[0][i])

        ## Final comosition step
        self.TR.Compose()

	## Discard submodel TR.rel to free memory
        for sm in self.submodels:
            sm.TR[0].rel= mgr.One()
       
        ## Top level hold relation
        ## This relation is allows no change in state
        self.hold = self.TR.vecps.HoldTR(self.TR.vecns)

        ## Add pipeline type constraints or else use tags
        setTR = []

        ## These are constraints between iterates
        ## This particular for loop requires the all sense operand MA to form a dependency loop. i.e. in-order single-issue model 
        for i in range(0,len(self.submodels)-1):
            Implication(setTR,self.submodels[i+1].sense.required[0][0],self.submodels[i].sense.known[0][0].P)
            Implication(setTR,self.submodels[i+1].sense.required[0][1],self.submodels[i].sense.known[0][1].P)
            Implication(setTR,self.submodels[i].sense.forget[0][0],(self.submodels[i+1].sense.accepted[0][0].P + self.submodels[i+1].sense.accepted[0][0].N))
            Implication(setTR,self.submodels[i].sense.forget[0][1],(self.submodels[i+1].sense.accepted[0][1].P + self.submodels[i+1].sense.accepted[0][1].N))

        ## Looping constraint
        ## This completes the cyclic loop with inverted, inter iteration dependencies from the last iterate's sense operand MA to the first iterate's mso
        Implication(setTR,self.submodels[0].sense.required[0][0],self.submodels[len(self.submodels)-1].sense.known[0][1].P)
        Implication(setTR,self.submodels[0].sense.required[0][1],self.submodels[len(self.submodels)-1].sense.known[0][0].P)

        Implication(setTR,self.submodels[len(self.submodels)-1].sense.forget[0][1], (self.submodels[0].sense.accepted[0][0].P + self.submodels[0].sense.accepted[0][0].N))
        Implication(setTR,self.submodels[len(self.submodels)-1].sense.forget[0][0], (self.submodels[0].sense.accepted[0][1].P + self.submodels[0].sense.accepted[0][1].N))

        ## Prune the appropriate relation through intersection
        vec = List2Vec(setTR)
        self.TR[0].rel = self.TR[0].rel & vec.And()

    ## This builds the control phase TR for the top-level composition
    def BuildcTR(self):

        self.cTR = Trans(1)
        self.cTR.PreCompose()

        for sm in self.submodels:
            self.cTR.Add(sm.cTR)

        self.cTR.Compose()

        ## Add pipeline type constraints

    ## Build resource use groupings
    def BuildGroups(self):
        
        allbounds = mgr.One()

        alu_grp = []
        cond_grp = []
        mult_grp = []
        table_grp = []
        mem_grp = []
        for sm in self.submodels:
            alu_grp = alu_grp + sm.alu_grp
            cond_grp = cond_grp + sm.cond_grp
            mult_grp = mult_grp + sm.mult_grp
            table_grp = table_grp + sm.table_grp
            mem_grp = mem_grp + sm.mem_grp

        ## Group all alu uses and build a concurrency constraint that allows 0 to ialu uses 
        self.alu_res = Group(0, ialu, alu_grp)
        ## Collect this constraint in the allbounds term
        allbounds = allbounds & self.alu_res.BuildActive()

        self.table_res = Group(0, 1, table_grp)      
        allbounds = allbounds & self.table_res.BuildActive()

        self.mult_res = Group(0, imult, mult_grp)
        allbounds = allbounds & self.mult_res.BuildActive()

        self.cond_res = Group(0, 1, cond_grp)      
        allbounds = allbounds & self.cond_res.BuildActive()

        ## live register concurrency constraints may be added to
        #self.mem_res = Group(0, 4, mem_grp)
        #allbounds = allbounds & self.mem_res.BuildActive()

        ## Prune the TR relation with allbounds
        self.TR.PruneT(allbounds, (0,1) )

## Here is the script that does all the high-level work

## Build a compostition of rotor iterates with iiterates (integer)
bm = Top(0,iiterates)

## Build the TR and cTR
bm.BuildTR()
bm.BuildcTR()
print "Top TR Info"
bm.TR.Info()
#bm.cTR.Collapse()
bm.cTR.Info()

## Set the variable mapping from ps to ns
bm.TR.SetVarMap()

## Build some required structures
bm.BuildVectors((0,0),(0,1)) # Don't flip guards

#bm.Reorder()
#bm.TR.Info()
#bm.cTR.Info()

## Do the viability prune
rs = bm.DualReachTRcTR(bm.taskstart[0])
bm.TR.PruneS(rs,(0,1))

## Collapse any paritions down to one
bm.cTR.Collapse()
bm.TR.Info()
#bm.cTR.PruneS(rs)
bm.cTR.Info()

## Do ROBDD reordering
bm.Reorder()
bm.TR.Info()
bm.cTR.Info()

## Apply resource concurrency constraints if ialu is not -1
if not (ialu == -1):
    print "Building groups",ialu,imult
    bm.BuildGroups()
    bm.TR.Info()

#bm.Reorder()
#bm.TR.Info()
#bm.cTR.Info()

#bm.pin = bm.p1.s1
## Specify the behavior cut MA
bm.pin = bm.submodels[0].sense

## Call the scheduling routine - CSchedule for CDFG scheduling
result = bm.CSchedule()

## execute the particular control prioritization script
execfile(mix)


print os.times()






















