#!/usr/bin/python -u
from malib import *
import sys
import os

outname = sys.argv[1]

if not outname == "term":
        print "Output to",outname
        sys.stdout = open(outname,"w")

ialu = int(sys.argv[2])
iiterates = int(sys.argv[3])
mix = sys.argv[4]

print os.times()

class Rotor(Model):

    def __init__(self,parent):
        Model.__init__(self,parent)
        self.submodels = []

        # Modeling NFA, BDD constructed as ordered here

        self.sense = COp1( self )
        
        self.l5 = COp1( self ) 
        self.s4 = COp1( self ) 
        self.s3 = COp1( self )
        self.s2 = COp1( self ) 
        self.l2 = COp1( self ) 
        self.c1  = COp1( self )
        self.c1c = Case( self, 2 )
        self.l1 = COp1( self ) 
        self.l3 = COp1( self ) 
        self.c2 = COp1( self )
        self.c2c = Case( self, 2 )
        self.s1 = COp1( self ) 
        self.s6 = COp1( self ) 
        self.l6 = COp1( self ) 
        self.s7 = COp1( self ) 
        self.s8 = COp1( self ) 
        self.l7 = COp1( self ) 
        self.c3 = COp1( self )
        self.c3c = Case( self, 2 )
        self.l9 = COp1( self ) 
        self.s11 = COp1( self ) 
        self.s5 = COp1( self )    
        self.s10 = COp1( self ) 
        self.l8 = COp1( self ) 
        self.s9 = COp1( self )   
        self.m1 = COp1( self )
        self.m2 = COp1( self )
        self.a1 = COp1( self )   
        self.m3 = COp1( self ) 
        self.m4 = COp1( self ) 
        self.a2 = COp1( self ) 

        ## For debugging
        self.AddNames()

    def BuildOperandLinks(self):

        self.sense.oinputs = ( (), )

        self.s1.oinputs = ( (), )
        self.c1.oinputs = ( ( DT(self.s1, 0, 0), ), )
        self.s2.oinputs = ( (), )
        self.c2.oinputs = ( ( DT(self.s2, 0, 0), ), )
        self.l1.oinputs = ( (), )
        self.l2.oinputs = ( ( DT(self.s2, 0, 0), ), )
        self.l3.oinputs = ( ( DT(self.s1, 0, 0), ), )
        self.s3.oinputs = ( ( DT(self.s2, 0, 0), ), )
        self.l5.oinputs = ( ( DT(self.s3, 0, 0), ), )
        self.s4.oinputs = ( ( DT(self.l5, 0, 0), ), )
        self.s5.oinputs = ( (), )
        self.c3.oinputs = ( ( DT(self.s5, 0, 0), ), )
        self.s6.oinputs = ( ( DT(self.s1, 0, 0), ), )
        self.l6.oinputs = ( ( DT(self.s6, 0, 0), ), )
        self.s7.oinputs = ( ( DT(self.l6, 0, 0), ), )
        self.l7.oinputs = ( ( DT(self.s5, 0, 0), ), )
        self.s8.oinputs = ( ( DT(self.l7, 0, 0), ), )
        self.s9.oinputs = ( (), )
        self.l8.oinputs = ( ( DT(self.s9, 0, 0), ), )
        self.s10.oinputs = ( ( DT(self.l8, 0, 0), ), )
        self.s11.oinputs = ( ( DT(self.s5, 0, 0), ), )
        self.l9.oinputs = ( ( DT(self.s11, 0, 0), ), )

        self.m1.oinputs = ( ( DT(self.l2, 0, 0), DT(self.s4, 0, 0), DT(self.s8, 0, 0), DT(self.l9, 0, 0) ), )
        self.m2.oinputs = ( ( DT(self.l1, 0, 0), DT(self.l3, 0, 0), DT(self.s7, 0, 0), DT(self.s10, 0, 0) ), )

        self.m4.oinputs = ( ( DT(self.l2, 0, 0), DT(self.s4, 0, 0), DT(self.s8, 0, 0), DT(self.l9, 0, 0) ), )
        self.m3.oinputs = ( ( DT(self.l1, 0, 0), DT(self.l3, 0, 0), DT(self.s7, 0, 0), DT(self.s10, 0, 0) ), )

        self.a1.oinputs = ( ( DT(self.m1, 0, 0), ),
                            ( DT(self.m2, 0, 0), ) )

        self.a2.oinputs = ( ( DT(self.m3, 0, 0), ),
                            ( DT(self.m4, 0, 0), ) )

        ## Control specific
        self.c1c.ooutputs = ( (), )
        self.c1c.oinputs = ( (), )

        self.c2c.ooutputs = ( (), )
        self.c2c.oinputs = ( (), )       

        self.c3c.ooutputs = ( (), )
        self.c3c.oinputs = ( (), )

    def BuildControlStructure(self):

        # Define forks
        self.forks = [self.c1c, self.c2c, self.c3c]
           
        set = []
		
        # Synchronize control automata with control operations
        self.c1c.Sync(set, self.c1, 0)
        self.c2c.Sync(set, self.c2, 0)
        self.c3c.Sync(set, self.c3, 0)
        
        vec = List2Vec(set)
        syncterm = vec.And()
        self.TR[0].rel = self.TR[0].rel & syncterm

        ## Assign tasks to control blocks
        ## specify ncb for global operations
        ncb = [State(-2), State(-2)]
        self.sense.ncb = ncb
        self.s1.ncb = ncb
        self.c1.ncb = ncb
        self.c1c.ncb = ncb
        self.m1.ncb = ncb
        self.m2.ncb = ncb
        self.m3.ncb = ncb
        self.m4.ncb = ncb
        self.a1.ncb = ncb
        self.a2.ncb = ncb

        self.c1c.OnlyWhen( (0, ), (self.s2, self.c2, self.c2c, self.l1, self.l2, self.l3, self.s3, self.l5, self.s4) )
        self.c1c.OnlyWhen( (1, ), (self.s5, self.c3, self.c3c, self.s6, self.l6, self.s7, self.s8, self.l7, self.s9, self.l8, self.s10, self.s11, self.l9) )
        
        self.c2c.OnlyWhen1( (0, ), (self.l1, self.l2) )
        self.c2c.OnlyWhen1( (1, ), (self.l3, self.s3, self.l5, self.s4) )

        self.c3c.OnlyWhen1( (0, ), (self.s6, self.l6, self.s7, self.l7, self.s8) )
        self.c3c.OnlyWhen1( (1, ), (self.s9, self.l8, self.s10, self.s11, self.l9) )

    def BuildTRcTR(self):

        self.BuildControlStructure()
        self.BuildOperandLinks()

        for sm in self.submodels:
            sm.BuildOOutputs()

        ss0 = self.taskstart[0][0]
        ss1 = self.taskstart[0][1]

        ss0.Smooth(self.sense.smth)
        ss1.Smooth(self.sense.smth)

        self.TR[0].rel = self.TR[0].rel & ( self.sense.required[0][0] >> (ss0.P | ss0.N) )
        self.TR[0].rel = self.TR[0].rel & ( self.sense.required[0][1] >> (ss1.P | ss1.N) )

        setTR  = []
        setcTR = []

        ## Insist on one sense (allowed odd/even iterations depenend on instruction sense)
        for sm in self.submodels:
            if not (sm == self.sense):
                setTR.append( (sm.taskstart[0][0].P & ~sm.taskstart[0][0].N) >> self.sense.known[0][0].P )
                setTR.append( (sm.taskstart[0][1].P & ~sm.taskstart[0][1].N) >> self.sense.known[0][1].P )
            else:
                print "Found sense"

          
        for sm in self.submodels:
            sm.BuildTRcTRConstraints(setTR, setcTR)

        vec = List2Vec(setTR)
        self.TR[0].rel = self.TR[0].rel & vec.And()

        vec = List2Vec(setcTR)
        self.cTR[0].rel = self.cTR[0].rel & vec.And()

        ## Top level hold relation
        self.hold = self.TR.vecps.HoldTR(self.TR.vecns)


    def BuildComplete(self,base):

        print "Complete Build"
        self.BuildCompositeState(base)
        Model.BuildCompositeTR(self)
        Model.BuildCompositecTR(self)
        #self.BuildOperandLinks()
        #self.BuildControlStructure()
        print "Initial CMA TR and cTR:"
        self.TR.Info()
        self.cTR.Info()

        self.TR.SetVarMap()
        self.BuildVectors((0,0),(0,1)) # Don't flip guards
        self.BuildTRcTR()
        print "Dependency Refined CMA TR and cTR:"
        self.TR.Info()
        self.cTR.Info()

        self.DefineGroups()

        rs = self.DualReachTRcTR(self.taskstart[0])
        self.TR.PruneS(rs)
        ##self.cTR.PruneS(rs)
        print "Reachable Refined CMA TR and cTR:"     
        self.TR.Info()
        self.cTR.Info()
        
    def BuildShifted(self,base,original):

        print "Shifted Build"
        self.BuildCompositeState(base)

        self.TR = Trans(0,original.TR)
        self.cTR = Trans(0,original.cTR)

        shft = ShiftBDD(original.base,original.width+original.base,base)
        self.TR.Shift(shft)
        self.cTR.Shift(shft)

        self.ss = []
        for ss in original.ss:
	        ## Copy state
            temp = State(-1,ss)
            temp.Shift(shft)
            self.ss.append(temp)

        ## Shift memory care conditions
        i = 0
        for sm in self.submodels:
            try:
                sm.memcare[0][0] = shft.Shift(original.submodels[i].memcare[0][0])
                sm.memcare[0][1] = shft.Shift(original.submodels[i].memcare[0][1])
            except AttributeError:
                print "No memorycare for",sm.name
            i = i + 1
        
        ## Define Forks
        self.forks = [self.c1c, self.c2c, self.c3c]
        self.DefineGroups()

    def DefineGroups(self):

        ## Define Groups
        self.alu_grp = [self.s1.busy[0], self.s2.busy[0], self.s3.busy[0], self.s4.busy[0], self.s5.busy[0], self.s6.busy[0], self.s7.busy[0], self.s8.busy[0], self.s9.busy[0], self.s10.busy[0], self.s11.busy[0], self.a1.busy[0], self.a2.busy[0] ]

        self.cond_grp = [ self.c1.busy[0], self.c2.busy[0], self.c3.busy[0] ]
        
        self.table_grp = [ self.l1.busy[0], self.l2.busy[0], self.l3.busy[0], self.l5.busy[0], self.l6.busy[0], self.l7.busy[0], self.l8.busy[0], self.l9.busy[0] ]

        self.mult_grp = [ self.m1.busy[0], self.m2.busy[0], self.m3.busy[0], self.m4.busy[0] ]

        self.mem_grp = [ self.s1.memcare[0], self.s2.memcare[0], self.s3.memcare[0], self.s4.memcare[0], self.s5.memcare[0], self.s6.memcare[0], self.s7.memcare[0], self.s8.memcare[0], self.s9.memcare[0], self.s10.memcare[0], self.s11.memcare[0], self.l1.memcare[0], self.l2.memcare[0], self.l3.memcare[0], self.l5.memcare[0], self.l6.memcare[0], self.l7.memcare[0], self.l8.memcare[0], self.l9.memcare[0], self.m1.memcare[0], self.m2.memcare[0], self.m3.memcare[0], self.m4.memcare[0] ]

class Top(Model):

    def __init__(self,parent,iterates):
		
        Model.__init__(self,parent)

        ## Pipe copies
        self.submodels = []

        self.submodels.append(Rotor(0))
        self.submodels[0].BuildComplete(0)
        
        for i in range(1,iterates):
            self.submodels.append(Rotor(0))
            base = self.submodels[i-1].width+self.submodels[i-1].base
            self.submodels[i].BuildShifted(base,self.submodels[0])
        
        self.base = 0
        self.width = self.submodels[iterates-1].base+self.submodels[iterates-1].width

    def BuildTR(self):

        self.forks = []
        for sm in self.submodels:
            self.forks = self.forks + sm.forks

	## Build resolve labeled state
        vec = DdArray(len(self.forks))
        self.resolve = State(-2)
        for frk in self.forks:
            self.resolve = self.resolve | frk.resolve

        self.TR = Trans(1) 
        self.TR.PreCompose()

        self.taskstart = ([State(0),State(0)], )

        for sm in self.submodels:
            self.TR.Add(sm.TR)
            for i in range(0,len(sm.taskstart[0])):
                try:
                    self.taskstart[0][i] = self.taskstart[0][i] & sm.taskstart[0][i]
                except IndexError:
                    self.taskstart[0].append(sm.taskstart[0][i])
          
        self.TR.Compose()

	## Discard submodel TR.rel to make room
        for sm in self.submodels:
            sm.TR[0].rel= mgr.One()
       
        ## Top level hold relation
        self.hold = self.TR.vecps.HoldTR(self.TR.vecns)

        ## Add pipeline type constraints or else use tags
        setTR = []

        for i in range(0,len(self.submodels)-1):
            Implication(setTR,self.submodels[i+1].sense.required[0][0],self.submodels[i].sense.known[0][0].P)
            Implication(setTR,self.submodels[i+1].sense.required[0][1],self.submodels[i].sense.known[0][1].P)
            Implication(setTR,self.submodels[i].sense.forget[0][0],(self.submodels[i+1].sense.accepted[0][0].P + self.submodels[i+1].sense.accepted[0][0].N))
            Implication(setTR,self.submodels[i].sense.forget[0][1],(self.submodels[i+1].sense.accepted[0][1].P + self.submodels[i+1].sense.accepted[0][1].N))

        ## Looping constraint
        Implication(setTR,self.submodels[0].sense.required[0][0],self.submodels[len(self.submodels)-1].sense.known[0][1].P)
        Implication(setTR,self.submodels[0].sense.required[0][1],self.submodels[len(self.submodels)-1].sense.known[0][0].P)

        Implication(setTR,self.submodels[len(self.submodels)-1].sense.forget[0][1], (self.submodels[0].sense.accepted[0][0].P + self.submodels[0].sense.accepted[0][0].N))
        Implication(setTR,self.submodels[len(self.submodels)-1].sense.forget[0][0], (self.submodels[0].sense.accepted[0][1].P + self.submodels[0].sense.accepted[0][1].N))

        vec = List2Vec(setTR)
        self.TR[0].rel = self.TR[0].rel & vec.And()


    def BuildcTR(self):

        self.cTR = Trans(1)
        self.cTR.PreCompose()

        for sm in self.submodels:
            self.cTR.Add(sm.cTR)

        self.cTR.Compose()

        ## Add pipeline type constraints

    def BuildGroups(self):
        
        allbounds = mgr.One()

        alu_grp = []
        cond_grp = []
        mult_grp = []
        table_grp = []
        mem_grp = []
        for sm in self.submodels:
            alu_grp = alu_grp + sm.alu_grp
            cond_grp = cond_grp + sm.cond_grp
            mult_grp = mult_grp + sm.mult_grp
            table_grp = table_grp + sm.table_grp
            mem_grp = mem_grp + sm.mem_grp

        alu_grp = alu_grp + mult_grp

        self.alu_res = Group(0, ialu, alu_grp)
        allbounds = allbounds & self.alu_res.BuildActive()

        self.table_res = Group(0, 1, table_grp)      
        allbounds = allbounds & self.table_res.BuildActive()

        #self.mult_res = Group(0, imult, mult_grp)
        #allbounds = allbounds & self.mult_res.BuildActive()

        self.cond_res = Group(0, 1, cond_grp)      
        allbounds = allbounds & self.cond_res.BuildActive()

        #self.mem_res = Group(0, 4, mem_grp)
        #allbounds = allbounds & self.mem_res.BuildActive()

        self.TR.PruneT(allbounds, (0,1) )

bm = Top(0,iiterates)

bm.BuildTR()
bm.BuildcTR()
print "Top TR Info"
bm.TR.Info()
#bm.cTR.Collapse()
bm.cTR.Info()

bm.TR.SetVarMap()
bm.BuildVectors((0,0),(0,1)) # Don't flip guards

#bm.Reorder()
#bm.TR.Info()
#bm.cTR.Info()

rs = bm.DualReachTRcTR(bm.taskstart[0])
	
bm.TR.PruneS(rs,(0,1))
bm.cTR.Collapse()
bm.TR.Info()
#bm.cTR.PruneS(rs)
bm.cTR.Info()

bm.Reorder()
bm.TR.Info()
bm.cTR.Info()

if not (ialu == -1):
    print "Building groups",ialu
    bm.BuildGroups()
    bm.TR.Info()

#bm.Reorder()
#bm.TR.Info()
#bm.cTR.Info()

#bm.pin = bm.p1.s1
bm.pin = bm.submodels[0].sense
result = bm.CSchedule()

execfile(mix)


print os.times()

