#!/usr/bin/python -u
## Change the above to match your site
## The option -u is for unbuffered output

## Import required libraries
import sys
import os
from malib import *

## To run this file, use the command line options:
## ewf-1.py term alus mults registers
## where term is the word term and dumps output to the terminal
## where alus is number of alus (an integer)
## where mults is the number of multipliers (an integer)
## where registers is the number of registers (an integer)
## A good initial trial is
## ewf-1.py term 2 1 100
## 100 is a high enough number of registers that no register contraint will result

## This particular version of EWF originated with Ivan

outname = sys.argv[1]

if not outname == "term":
        print "Output to",outname
        sys.stdout = open(outname,"w")

alus = int(sys.argv[2])
mults = int(sys.argv[3])
registers = int(sys.argv[4])

print os.times()



## A CMA (refer to my dissretation) is built for each benchmark
## This corresponds to a python class derived from the base model class in ma.py
class Top(Model):


        ## The constructs the model
	def __init__(self,parent):

                ## Must manually call the base constructor
		Model.__init__(self,parent)

                ## This is the list of all submodels (MA) in this CMA
		self.submodels = []

                ## self.a0 is a model of type COp1 where COp1 is defined in malib.py
                ## the 1 or 0 indicates which partition to build this MA in and is purely for
                ## bdd optmization and does not change the scheduling problem or results
		self.a0 = COp1(self,1)
		self.a1 = COp1(self,1)
		self.a2 = COp1(self,1)
		self.a3 = COp1(self,0)
		self.a4 = COp1(self,0)
		self.m5 = COp2p(self,0)
		self.a6 = COp1(self,0)
		self.a7 = COp1(self,0)
		self.a8 = COp1(self,0)
		self.a9 = COp1(self,1)
		self.a10 = COp1(self,0)
		self.m11 = COp2p(self,0)
		self.a12 = COp1(self,0)
		self.a13 = COp1(self,1)
		self.a14 = COp1(self,0)
		self.m15 = COp2p(self,0)
		self.a16 = COp1(self,1)
		self.a17 = COp1(self,1)
		self.m18 = COp2p(self,1)
		self.a19 = COp1(self,0)
		self.m20 = COp2p(self,0)
		self.a21 = COp1(self,0)
		self.m22 = COp2p(self,0)
		self.a23 = COp1(self,0)
		self.a24 = COp1(self,0)
		self.a25 = COp1(self,0)
		self.a26 = COp1(self,0)
		self.m27 = COp2p(self,0)
		self.a28 = COp1(self,0)
		self.a29 = COp1(self,1)
		self.a30 = COp1(self,1)
		self.m31 = COp2p(self,0)
		self.a32 = COp1(self,0)
		self.a33 = COp1(self,1)

                ## This is a list of all MA (tasks within the composition) that produce operands which are visible are required external.
		self.outtags = [self.a8, self.a17, self.m20, self.a21, self.a28, self.a29, self.a32, self.a33, ]
                self.AddNames()

        ## This function builds the dependency and capacity implications
	def BuildRelation(self):
                ## Build template framework
		Model.BuildCompositeTR(self)
		set = []

                ## IEnable(set,self.a28.known[0],self.a28.required[1]) means that there is an inverted edge (inter iteration dependency)
                ## from output operand index 0 of MA a28 to input operand index 1 of MA a28 (This MA depends on a result it produces!)
		IEnable(set,self.a28.known[0],self.a28.required[1])
		IEnable(set,self.a32.known[0],self.a32.required[1])

                ## Enable(set,self.a0.known[0],self.a1.required[0]) means that there is a normal edge (intra iteration dependency)
                ## from output operand index 0 of MA a0 to input operand index 0 of MA a1
		Enable(set,self.a0.known[0],self.a1.required[0])
		Enable(set,self.a2.known[0],self.a3.required[0])
		Enable(set,self.m5.known[0],self.a6.required[0])
		Enable(set,self.a6.known[0],self.a7.required[1])
		Enable(set,self.a7.known[0],self.a8.required[0])
		Enable(set,self.a10.known[0],self.m11.required[0])
		Enable(set,self.m11.known[0],self.a12.required[0])
		Enable(set,self.a13.known[0],self.a12.required[1])
		Enable(set,self.a13.known[0],self.a14.required[0])
		Enable(set,self.m15.known[0],self.a16.required[0])
		Enable(set,self.a16.known[0],self.a17.required[1])
		Enable(set,self.m18.known[0],self.a19.required[0])
		Enable(set,self.m20.known[0],self.a21.required[0])
		Enable(set,self.m22.known[0],self.a23.required[0])
		Enable(set,self.a23.known[0],self.a24.required[1])
		Enable(set,self.a26.known[0],self.m27.required[0])
		Enable(set,self.m27.known[0],self.a28.required[0])
		Enable(set,self.a28.known[0],self.a29.required[0])
		Enable(set,self.a30.known[0],self.m31.required[0])
		Enable(set,self.m31.known[0],self.a32.required[0])
		Enable(set,self.a32.known[0],self.a33.required[1])
		Enable(set,self.a0.known[0],self.a2.required[0])
		Enable(set,self.a2.known[0],self.a4.required[0])
		Enable(set,self.a12.known[0],self.a14.required[1])
		Enable(set,self.a19.known[0],self.a17.required[0])
		Enable(set,self.a23.known[0],self.a21.required[1])
		IEnable(set,self.a21.known[0],self.a23.required[1])
		Enable(set,self.a23.known[0],self.a25.required[0])
		Enable(set,self.a24.known[0],self.a26.required[0])
		IEnable(set,self.a28.known[0],self.a26.required[1])
		IEnable(set,self.a32.known[0],self.a30.required[1])
		Enable(set,self.a6.known[0],self.a3.required[1])
		Enable(set,self.a10.known[0],self.a7.required[0])
		Enable(set,self.a6.known[0],self.a9.required[1])
		Enable(set,self.a13.known[0],self.a10.required[1])
		Enable(set,self.a26.known[0],self.a29.required[1])
		Enable(set,self.a30.known[0],self.a33.required[0])
		IEnable(set,self.a8.known[0],self.a4.required[1])
		Enable(set,self.a2.known[0],self.a6.required[1])
		Enable(set,self.a12.known[0],self.a8.required[1])
		IEnable(set,self.a21.known[0],self.a25.required[1])
		Enable(set,self.a10.known[0],self.m5.required[0])
		Enable(set,self.a25.known[0],self.m20.required[0])
		Enable(set,self.a4.known[0],self.a10.required[0])
		Enable(set,self.a16.known[0],self.a9.required[0])
		IEnable(set,self.a21.known[0],self.a13.required[1])
		Enable(set,self.a14.known[0],self.m22.required[0])
		Enable(set,self.a3.known[0],self.m15.required[0])
		Enable(set,self.a12.known[0],self.a24.required[0])
		Enable(set,self.a16.known[0],self.a1.required[1])

                ## All the dependency edges are save in the list called set
                ## This is turned into a bdd vector structure
		vec = List2Vec(set)
                ## The dependencies are intersected to prune the transition relation
                ## TR the transition relation object described in pyfa.py
                ## TR[0] refers to partition 0
                ## TR[0].rel refers to the relaion BDD
		self.TR[0].rel = self.TR[0].rel & vec.And()

                ## Clear the set list
		set = []

                ## These dependencies go into the second partition
                ## This is the optimization described on page 77 of my dissertation
                ## as 'Long' CMA constraints
		IEnable(set,self.a29.known[0],self.a13.required[0])
		Enable(set,self.a0.known[0],self.a16.required[1])
		IEnable(set,self.a17.known[0],self.a0.required[0])
		Enable(set,self.a1.known[0],self.m18.required[0])
		Enable(set,self.a9.known[0],self.a30.required[0])
		IEnable(set,self.a33.known[0],self.a2.required[1])


		vec = List2Vec(set)
		self.TR[1].rel = self.TR[1].rel & vec.And()

                ## Clear the set list
		set = []


                ## Now build capacity constraints
                ## This is a repeat of all the Enable edges except the
                ## constraint is built in terms of known and accepted
		ICapacity(set,self.a28.known[0],self.a28.accepted[1])
		ICapacity(set,self.a32.known[0],self.a32.accepted[1])
		Capacity(set,self.a0.known[0],self.a1.accepted[0])
		Capacity(set,self.a2.known[0],self.a3.accepted[0])
		Capacity(set,self.m5.known[0],self.a6.accepted[0])
		Capacity(set,self.a6.known[0],self.a7.accepted[1])
		Capacity(set,self.a7.known[0],self.a8.accepted[0])
		Capacity(set,self.a10.known[0],self.m11.accepted[0])
		Capacity(set,self.m11.known[0],self.a12.accepted[0])
		Capacity(set,self.a13.known[0],self.a12.accepted[1])
		Capacity(set,self.a13.known[0],self.a14.accepted[0])
		Capacity(set,self.m15.known[0],self.a16.accepted[0])
		Capacity(set,self.a16.known[0],self.a17.accepted[1])
		Capacity(set,self.m18.known[0],self.a19.accepted[0])
		Capacity(set,self.m20.known[0],self.a21.accepted[0])
		Capacity(set,self.m22.known[0],self.a23.accepted[0])
		Capacity(set,self.a23.known[0],self.a24.accepted[1])
		Capacity(set,self.a26.known[0],self.m27.accepted[0])
		Capacity(set,self.m27.known[0],self.a28.accepted[0])
		Capacity(set,self.a28.known[0],self.a29.accepted[0])
		Capacity(set,self.a30.known[0],self.m31.accepted[0])
		Capacity(set,self.m31.known[0],self.a32.accepted[0])
		Capacity(set,self.a32.known[0],self.a33.accepted[1])
		Capacity(set,self.a0.known[0],self.a2.accepted[0])
		Capacity(set,self.a2.known[0],self.a4.accepted[0])
		Capacity(set,self.a12.known[0],self.a14.accepted[1])
		Capacity(set,self.a19.known[0],self.a17.accepted[0])
		Capacity(set,self.a23.known[0],self.a21.accepted[1])
		ICapacity(set,self.a21.known[0],self.a23.accepted[1])
		Capacity(set,self.a23.known[0],self.a25.accepted[0])
		Capacity(set,self.a24.known[0],self.a26.accepted[0])
		ICapacity(set,self.a28.known[0],self.a26.accepted[1])
		ICapacity(set,self.a32.known[0],self.a30.accepted[1])
		Capacity(set,self.a6.known[0],self.a3.accepted[1])
		Capacity(set,self.a10.known[0],self.a7.accepted[0])
		Capacity(set,self.a6.known[0],self.a9.accepted[1])
		Capacity(set,self.a13.known[0],self.a10.accepted[1])
		Capacity(set,self.a26.known[0],self.a29.accepted[1])
		Capacity(set,self.a30.known[0],self.a33.accepted[0])
		ICapacity(set,self.a8.known[0],self.a4.accepted[1])
		Capacity(set,self.a2.known[0],self.a6.accepted[1])
		Capacity(set,self.a12.known[0],self.a8.accepted[1])
		ICapacity(set,self.a21.known[0],self.a25.accepted[1])
		Capacity(set,self.a10.known[0],self.m5.accepted[0])
		Capacity(set,self.a25.known[0],self.m20.accepted[0])
		Capacity(set,self.a4.known[0],self.a10.accepted[0])
		Capacity(set,self.a16.known[0],self.a9.accepted[0])
		ICapacity(set,self.a21.known[0],self.a13.accepted[1])
		Capacity(set,self.a14.known[0],self.m22.accepted[0])
		Capacity(set,self.a3.known[0],self.m15.accepted[0])
		Capacity(set,self.a12.known[0],self.a24.accepted[0])
		Capacity(set,self.a16.known[0],self.a1.accepted[1])


		vec = List2Vec(set)
		self.TR[0].rel = self.TR[0].rel & vec.And()

                ## Again put long constraints in second partition
		set = []

		ICapacity(set,self.a29.known[0],self.a13.accepted[0])
		Capacity(set,self.a0.known[0],self.a16.accepted[1])
		ICapacity(set,self.a17.known[0],self.a0.accepted[0])
		Capacity(set,self.a1.known[0],self.m18.accepted[0])
		Capacity(set,self.a9.known[0],self.a30.accepted[0])
		ICapacity(set,self.a33.known[0],self.a2.accepted[1])


		vec = List2Vec(set)
		self.TR[1].rel = self.TR[1].rel & vec.And()

        ## This function builds concurrency constraints as described on pages 49 and 58 in the dissertation
	def BuildGroups(self):

                ## A set of all busy labels that represent the same resource are grouped in a list
                ## The index for busy indicates a particular resource
                alu_list = [self.a0.busy[0], self.a1.busy[0], self.a2.busy[0], self.a3.busy[0], self.a4.busy[0], self.a6.busy[0], self.a7.busy[0], self.a8.busy[0], self.a9.busy[0], self.a10.busy[0], self.a12.busy[0], self.a13.busy[0], self.a14.busy[0], self.a16.busy[0], self.a17.busy[0], self.a19.busy[0], self.a21.busy[0], self.a23.busy[0], self.a24.busy[0], self.a25.busy[0], self.a26.busy[0], self.a28.busy[0], self.a29.busy[0], self.a30.busy[0], self.a32.busy[0], self.a33.busy[0]]

                ## Create a group object
                ## 0 is the lower bound
                ## alus is the upper bound
                ## alu_list is the group of bdds to constrain
		self.alu_res = Group(0, alus, alu_list)
		self.TR.PruneT(self.alu_res.BuildActive(),(0,1))

                ## Impose the multiplier concurrency limits
                mult_list = [self.m5.busy[0], self.m11.busy[0], self.m15.busy[0], self.m18.busy[0], self.m20.busy[0], self.m22.busy[0], self.m27.busy[0], self.m31.busy[0]]
		self.mult_res = Group(0, mults, mult_list)
		self.TR.PruneT(self.mult_res.BuildActive(),(0,1))

        ## This function builds concurrency constraints for memory
	def BuildMemory(self):

                ## self.a28 in self.a28.BuildIMemory is the producer of the result
                ## The 0 is the particular index that we care about in memory
                ## The MA that accepts this operand is also a28. It is input operand 1 [1] that is accepted
                ## These correspond to the dependency edges (again....) I choose to break all edges into
                ## three separate applicaitons (Enable, Capacity, BuildMemory)
                ## The I represents an inverted (inter iteration dependency) edge
		self.a28.BuildIMemory(0,self.a28.accepted[1])
		self.a32.BuildIMemory(0,self.a32.accepted[1])
		self.a0.BuildMemory(0,self.a1.accepted[0])
		self.a2.BuildMemory(0,self.a3.accepted[0])
		self.m5.BuildMemory(0,self.a6.accepted[0])
		self.a6.BuildMemory(0,self.a7.accepted[1])
		self.a7.BuildMemory(0,self.a8.accepted[0])
		self.a10.BuildMemory(0,self.m11.accepted[0])
		self.m11.BuildMemory(0,self.a12.accepted[0])
		self.a13.BuildMemory(0,self.a12.accepted[1])
		self.a13.BuildMemory(0,self.a14.accepted[0])
		self.m15.BuildMemory(0,self.a16.accepted[0])
		self.a16.BuildMemory(0,self.a17.accepted[1])
		self.m18.BuildMemory(0,self.a19.accepted[0])
		self.m20.BuildMemory(0,self.a21.accepted[0])
		self.m22.BuildMemory(0,self.a23.accepted[0])
		self.a23.BuildMemory(0,self.a24.accepted[1])
		self.a26.BuildMemory(0,self.m27.accepted[0])
		self.m27.BuildMemory(0,self.a28.accepted[0])
		self.a28.BuildMemory(0,self.a29.accepted[0])
		self.a30.BuildMemory(0,self.m31.accepted[0])
		self.m31.BuildMemory(0,self.a32.accepted[0])
		self.a32.BuildMemory(0,self.a33.accepted[1])
		self.a0.BuildMemory(0,self.a2.accepted[0])
		self.a2.BuildMemory(0,self.a4.accepted[0])
		self.a12.BuildMemory(0,self.a14.accepted[1])
		self.a19.BuildMemory(0,self.a17.accepted[0])
		self.a23.BuildMemory(0,self.a21.accepted[1])
		self.a21.BuildIMemory(0,self.a23.accepted[1])
		self.a23.BuildMemory(0,self.a25.accepted[0])
		self.a24.BuildMemory(0,self.a26.accepted[0])
		self.a28.BuildIMemory(0,self.a26.accepted[1])
		self.a32.BuildIMemory(0,self.a30.accepted[1])
		self.a6.BuildMemory(0,self.a3.accepted[1])
		self.a10.BuildMemory(0,self.a7.accepted[0])
		self.a6.BuildMemory(0,self.a9.accepted[1])
		self.a13.BuildMemory(0,self.a10.accepted[1])
		self.a26.BuildMemory(0,self.a29.accepted[1])
		self.a30.BuildMemory(0,self.a33.accepted[0])
		self.a8.BuildIMemory(0,self.a4.accepted[1])
		self.a2.BuildMemory(0,self.a6.accepted[1])
		self.a12.BuildMemory(0,self.a8.accepted[1])
		self.a21.BuildIMemory(0,self.a25.accepted[1])
		self.a10.BuildMemory(0,self.m5.accepted[0])
		self.a25.BuildMemory(0,self.m20.accepted[0])
		self.a4.BuildMemory(0,self.a10.accepted[0])
		self.a16.BuildMemory(0,self.a9.accepted[0])
		self.a21.BuildIMemory(0,self.a13.accepted[1])
		self.a14.BuildMemory(0,self.m22.accepted[0])
		self.a3.BuildMemory(0,self.m15.accepted[0])
		self.a12.BuildMemory(0,self.a24.accepted[0])
		self.a16.BuildMemory(0,self.a1.accepted[1])
		self.a29.BuildIMemory(0,self.a13.accepted[0])
		self.a0.BuildMemory(0,self.a16.accepted[1])
		self.a17.BuildIMemory(0,self.a0.accepted[0])
		self.a1.BuildMemory(0,self.m18.accepted[0])
		self.a9.BuildMemory(0,self.a30.accepted[0])
		self.a33.BuildIMemory(0,self.a2.accepted[1])

                ## This creates the memory grouping
                ## All values are stored in one pool of registers since there is just one grouping: mem_list
                mem_list = [self.a0.memcare[0], self.a1.memcare[0], self.a2.memcare[0], self.a3.memcare[0], self.a4.memcare[0], self.m5.memcare[0], self.a6.memcare[0], self.a7.memcare[0], self.a8.memcare[0], self.a9.memcare[0], self.a10.memcare[0], self.m11.memcare[0], self.a12.memcare[0], self.a13.memcare[0], self.a14.memcare[0], self.m15.memcare[0], self.a16.memcare[0], self.a17.memcare[0], self.m18.memcare[0], self.a19.memcare[0], self.m20.memcare[0], self.a21.memcare[0], self.m22.memcare[0], self.a23.memcare[0], self.a24.memcare[0], self.a25.memcare[0], self.a26.memcare[0], self.m27.memcare[0], self.a28.memcare[0], self.a29.memcare[0], self.a30.memcare[0], self.m31.memcare[0], self.a32.memcare[0], self.a33.memcare[0]]

                ## Lower bound is 0, upper bound is registers
		self.mem_res = Group(0, registers, mem_list)

                ## Prune the transition relation, only partitions 0
		self.TR.PruneT(self.mem_res.BuildActive(),(0,1))



## This is where the actual scheduling takes place -- as a scrit at the end of the file!

## create a benchmark object
bm = Top(0)

## Build state for all member MA, start with BDD index 0 and build up
bm.BuildCompositeState(0)

## Build the transition relation
bm.BuildRelation()
## Print some transition information
bm.TR.Info()

## Set the default variable map to present state ps maps to next state ns
bm.TR.SetVarMap()

## Build some vectors that are needed -- look at the call in ma.py
bm.BuildVectors((0,0),(0,1))
bm.TR.Info()

## Viability prune as described on page 55 of the disertation
bm.TR.ReachabilityPrune(bm.taskstart[0],(0,1))

## Collapse the partitioned transition relation down to one partition
bm.TR.Collapse()
bm.TR.Info()

## Apply function unit concurrency constraints
bm.BuildGroups()
bm.TR.Info()

## Apply memory concurrency constraints
bm.BuildMemory()
bm.TR.Info()

## This is the loop cut MA as described on page 103 of the dissertation
bm.pin = bm.a10

## This scheduling algorithm finds a directly closed path set as described on page 107 of the dissertation
## tag depth is the number of MA to tag during each iteration of closure
bm.tagdepth = 12
#bm.Reorder()

## This calls a scheduling algorithm in ma.py
## DF is for data-flow examples
result = bm.DFSchedule()

## This finds a witness schedule
result.Trace()

## This prints out the witness schedule
result.PrintTrace()


print os.times()

## All done!













