#!/usr/bin/python -u

## Simple Scalar Model
## Hazard, 0==none, 1=rs, 2=rt
from malib import *
import sys
import os

outname = sys.argv[1]

if not outname == "term":
        print "Output to",outname
        sys.stdout = open(outname,"w")

iconfig = sys.argv[4]
iiterates = int(sys.argv[3])
mix = sys.argv[2]

print os.times()

class RiscPipeStage(Model):

	def __init__(self,parent):

	    Model.__init__(self,parent)
	    self.submodels = []

	    ## Redundant operand indicators -- used to simplify interfacing
	    ## Maybe safely overlaid with pcout, instout, bypass1out from previous stage
	    self.pcin = Rename( self )    # This is ppc, need ppc too
	    self.instin = Rename( self )
	    self.bypassin = Rename( self )

	    ## Case statements
	    self.decode = COp1( self )
	    self.decodec = Case( self, 14)
	    self.blockc = Case( self, 2, self.decodec)
	    self.hazardc = Case( self, 3, self.decodec)

	    self.bopc = Case( self, 2 )   # Multivalue machine for branch taken/ not taken
	    self.bop = COp1( self )       # ALU operation to determine branch taken/not taken

	    ## pc preincrement and inst prefecth
	    self.insts   = COp2( self )   # Speculative instruction fetch
	    self.pcincrs = COp1( self )   # Speculative pc
	    
	    self.pcload = COp1( self )    # Operation to load new pc value as in jmp
	    
	    self.pcadj  = COp1( self )    # Adjust incremented pc for branches, requires ALU 
	    self.inst   = COp2( self )    # Instruction fetch based on incremented pc
	    self.pcincr = COp1( self )    # Basic program counter increment, requires ALU

	    ## integer instructions

	    ## Register file reads
	    self.wrhl = COp1( self )      # For to hilo
	    self.rdrs = COp1( self )      # Read rs
	    
	    
	    self.rdrt = COp1( self )      # Read rt	    
	   
	    self.rdhl = COp1( self )      # Read architecture hi/lo

	    ## Simple integer computations
	    self.wrrd = COp1( self )      # Writeback to rd
	    self.intus1 = COp1( self )    # For regreg1
	   

	    ## Complex integer computations
	    self.intuc = COp2ps( self )     # For regreg2
	    self.wrlo = COp1( self )      # For regreg2

	    ## Writebacks
	    self.wrrt = COp1( self )      # Writeback for rt

            ## Load/store address computation
	    self.memaddr = COp1( self )
	   
	    ## Load/store specific
	    self.memwr = COp2( self )
	    self.memrd = COp2( self )

	    self.meminc = COp1( self )

	    self.rdrf = COp1( self )	    

	    ## Redundant operand indicators -- used to simplify interfacing
	    self.pcout = Rename( self )
	    self.instout = Rename( self )
	    self.bypassout = Rename( self )

	    self.sense = self.instin

	    self.AddNames()


	def BuildOperandLinks(self):

	    ## Define task inputs -- Ordered: operand 0, 1, 2, 3, 4. Producer and Producer's operand index is included
	    ## Tuples are (producing task, operand index, operand sense)
	    byrs = self.hazardc.case[1]
	    byrt = self.hazardc.case[2]
	    
	    self.pcin.oinputs = ( )
	    self.instin.oinputs = ( )
	    self.bypassin.oinputs = ( )

	    self.decode.oinputs = ( ( DT(self.instin, 0, 0), ), ) 

	    self.insts.oinputs = ( ( DT(self.pcin, 0, 0), ), )
	    self.pcincrs.oinputs =  ( ( DT(self.pcin, 0, 0), ), )


	    ## rdrs is only applicable when decodec = case 7 or 8
	    temp = (self.decodec.case[7] | self.decodec.case[8])
	    cb0 = self.rdrs.icb[0] & temp
	    cb1 = self.rdrs.icb[1] & temp

	    temp = temp & self.hazardc.case[1]
	    by1_0 = self.decodec.syncedto.known[0][0] & temp
	    by1_1 = self.decodec.syncedto.known[0][1] & temp

	    self.inst.oinputs = ( ( DT(self.pcload, 0, 0), DT(self.pcadj, 0, 0), DT(self.rdrs, 0, 0, (cb0, cb1)), DT(self.bypassin, 0, 0, (by1_0, by1_1)) ), )
	    self.pcincr.oinputs = ( ( DT(self.pcload, 0, 0), DT(self.pcadj, 0, 0), DT(self.rdrs, 0, 0, (cb0, cb1)),  DT(self.bypassin, 0, 0, (by1_0, by1_1)) ), )  
	    
	    self.pcadj.oinputs = ( ( DT(self.pcin, 0, 0), ),
				   ( DT(self.instin, 0, 0), ) )

	    ## PCLOAD should really depend on pc not ppc
	    self.pcload.oinputs = ( ( DT(self.instin, 0, 0), ),
				    ( DT(self.pcin, 0, 0), ) )

	    ## decode is control
	    self.bop.oinputs = ( ( DT(self.rdrs, 0, 0), DT(self.bypassin, 0, 0, (byrs, byrs)) ),
				 ( DT(self.rdrt, 0, 0), DT(self.bypassin, 0, 0, (byrt, byrt)) ),
				 ( DT(self.decode, 0, 0), ) )
	    
	    ## Integer instructions
	    self.rdrs.oinputs = ( ( DT(self.instin, 0, 0), ), )
	    self.rdrt.oinputs = ( ( DT(self.instin, 0, 0), ), )
	    self.rdhl.oinputs = ( ( DT(self.decode, 0, 0), ), )

	    t0 = self.decodec.case[0]
	    t2 = self.decodec.case[2]
	    self.intus1.oinputs = ( ( DT(self.rdrs, 0, 0), DT(self.bypassin, 0, 0, (byrs, byrs)) ),
				    ( DT(self.rdrt, 0, 0, (t0, t0)), DT(self.bypassin, 0, 0, (t0 & byrt, t0 & byrt)), DT(self.instin, 0, 0, (t2,t2))),
				    ( DT(self.decode, 0, 0), ) )


	    self.intuc.oinputs  = ( ( DT(self.rdrs, 0, 0), DT(self.bypassin, 0, 0, (byrs, byrs)) ),
				    ( DT(self.rdrt, 0, 0), DT(self.bypassin, 0, 0, (byrt, byrt)) ),
				    ( DT(self.decode, 0, 0), ) )

	    ## Still needs decode or inst?
	    t0 = ~self.decodec.case[11]
	    t1 = self.decodec.case[0]
	    t2 = self.decodec.case[6] | self.decodec.case[8]
	    self.wrrd.oinputs   = ( ( DT(self.instin, 0, 0, (t0, t0)), DT(self.meminc, 0, 0) ), 
				    ( DT(self.pcin, 0, 0, (t2,t2)), DT(self.memrd, 1, 0), DT(self.rdhl, 0, 0), DT(self.intus1, 0, 0, (t1,t1)), ),
				    ( DT(self.decode, 0, 0), ) )

	    self.wrrt.oinputs   = ( ( DT(self.instin, 0, 0), ),
				    ( DT(self.memrd, 0, 0), DT(self.intus1, 0, 0), ) )

	    temp = self.decodec.case[3]
	    self.wrhl.oinputs   = ( ( DT(self.decode, 0, 0), ),
				    ( DT(self.intuc, 0, 0), DT(self.rdrs, 0, 0, (temp,temp)), DT(self.bypassin, 0, 0, (temp & byrs, temp & byrs)) ) )

	    self.wrlo.oinputs   = ( ( DT(self.intuc, 1, 0), ), )

	    ## Load/store specific 
	    self.rdrf.oinputs   = ( ( DT(self.meminc, 0, 0), ), )

	    self.meminc.oinputs = ( ( DT(self.instin, 0, 0), ), )

	    ## memaddr is only applicable during all memory address computations
	    self.memrd.oinputs = ( ( DT(self.memaddr, 0, 0), ), )

	    self.memwr.oinputs = ( ( DT(self.memaddr, 0, 0), ),
				   ( DT(self.rdrt, 0, 0), DT(self.bypassin, 0, 0, (byrt, byrt)) ),
				   ( DT(self.rdrf, 0, 0), ) )

	    ## Load/store address computation
	    self.memaddr.oinputs = ( ( DT(self.instin, 0, 0), ),
				     ( DT(self.bypassin, 0, 0, (byrs, byrs)), DT(self.rdrs, 0, 0) ) )

	    ## Operand renaming
	    self.pcout.oinputs = ( ( DT(self.pcincr, 0, 0), DT(self.pcincrs, 0, 0) ), )

	    b1a = self.blockc.case[1] & self.blockc.syncedto.known[0][0]
	    b1b = self.blockc.case[1] & self.blockc.syncedto.known[0][1]
	    b0 = self.blockc.case[0]

	    self.instout.oinputs = ( ( DT(self.insts, 0, 0), DT(self.inst, 0, 0),
				       DT(self.wrrd, 0, 0, (b1a, b1b)),
				       DT(self.wrrt, 0, 0, (b1a, b1b)),
				       DT(self.wrhl, 0, 0, (b1a, b1b)),
				       DT(self.wrlo, 0, 0, (b1a, b1b))), )

	    t2 = self.decodec.case[6] | self.decodec.case[8]
	    temp0 = self.decodec.syncedto.known[0][0] & t2
	    temp1 = self.decodec.syncedto.known[0][1] & t2
	    t11 = self.decodec.case[11]
	    self.bypassout.oinputs = ( ( DT(self.rdhl, 0, 0), DT(self.intus1, 0, 0),
					 DT(self.pcin, 0, 0, (temp0, temp1)),
					 DT(self.memrd, 0, 0),
					 DT(self.memrd, 1, 0, (t11,t11))), )
	    
	    ## New versus old
	    ## Control capacities -- written on outputs
	    self.bopc.ooutputs = ( (), )
	    self.bopc.oinputs = ( (), )

	    self.decodec.ooutputs = ( (), )
	    self.decodec.oinputs = ( (), )

	    self.blockc.ooutputs = ( (), )
	    self.blockc.oinputs = ( (), )

	    self.hazardc.ooutputs = ( (), )
	    self.hazardc.oinputs = ( (), )
	    
	def BuildTRcTR(self):

	    self.BuildControlStructure()
	    self.BuildOperandLinks()

	    for sm in self.submodels:
	         sm.BuildOOutputs()

	    setTR  = []
	    setcTR = []

	    ## Insist on single instruction in instruction execution task
	    ## Exclude interface input machines
	    ss0 = self.taskstart[0][0]
	    ss1 = self.taskstart[0][1]

	    #for sm in (self.instin, self.pcin, self.instout, self.pcout):
	    for sm in (self.bypassin, self.instin, self.pcin):
	       ss0.Smooth(sm.smth)
	       ss1.Smooth(sm.smth)

	    ## In cTR since that is where the rename machine lives
	    self.cTR[0].rel = self.cTR[0].rel & ( (self.instin.required[0][0] >> ss0.P) & (self.instin.required[0][1] >> ss1.P) )

	    ## Insist on one sense (allowed odd/even iterations depenend on instruction sense)
	    for sm in self.submodels:
	        if not (sm == self.instin or sm == self.pcin or sm == self.bypassin):
		     setTR.append( (sm.taskstart[0][0].P & ~sm.taskstart[0][0].N) >> self.instin.known[0][0].P )
		     setTR.append( (sm.taskstart[0][1].P & ~sm.taskstart[0][1].N) >> self.instin.known[0][1].P )
		else:
		     print "Found instin, pcin or bypassin"

	    ## Build input enabling terms
	    for sm in self.submodels:
	         sm.BuildTRcTRConstraints(setTR, setcTR)

	    vec = List2Vec(setTR)
	    self.TR[0].rel = self.TR[0].rel & vec.And()

	    vec = List2Vec(setcTR)
	    self.cTR[0].rel = self.cTR[0].rel & vec.And()
	     	    
	def BuildControlStructure(self):

            # Define forks
            self.forks = [ self.bopc, self.hazardc, self.blockc, self.decodec ]

	    set = []
		
            # Synchronize control automata with control operations        # Synchronize control automata with control operations
	    self.decodec.Sync( set, self.decode,0)
	    self.hazardc.Sync( set, self.decode,0)
	    self.blockc.Sync( set, self.decode,0)
	    self.bopc.Sync( set, self.bop,0)
	
            vec = List2Vec(set)
	    syncterm = vec.And()
            self.TR[0].rel = self.TR[0].rel & syncterm

            ## Assign Operations to specific cases

	    ## specify ncb for global operations
	    ncb = [State(-2), State(-2)]
	    self.decodec.ncb = ncb
	    self.hazardc.ncb = ncb
	    self.blockc.ncb = ncb
	    self.decode.ncb = ncb
	    self.pcin.ncb = ncb
	    self.pcout.ncb = ncb
	    self.instin.ncb = ncb
	    self.instout.ncb = ncb
	    self.bypassin.ncb = ncb
	    ##self.bypassout.ncb = ncb


	    ## Speculate instruction prefetch and pc preincrement
	    self.decodec.OnlyWhen( (0, 1, 2, 3, 4, 10, 11, 12, 13), (self.insts, self.pcincrs), (9, ) )

	    ## Additional information for branch case not taken
	    cb0 = ( (self.decodec.syncedto.known[0][0] & self.decodec.case[9]) &
		    (self.bopc.syncedto.known[0][0] & self.bopc.case[0]) )
	    cb1 = ( (self.decodec.syncedto.known[0][1] & self.decodec.case[9]) &
		    (self.bopc.syncedto.known[0][1] & self.bopc.case[0]) )
 
	    ncb0 = ( (self.decodec.syncedto.known[0][0] & self.decodec.case[9]) &
		    (self.bopc.syncedto.known[0][0] & ~self.bopc.case[0]) )
	    ncb1 = ( (self.decodec.syncedto.known[0][1] & self.decodec.case[9]) &
		    (self.bopc.syncedto.known[0][1] & ~self.bopc.case[0]) )

	    self.insts.AddCB(cb0,cb1,ncb0,ncb1)
	    self.pcincrs.AddCB(cb0,cb1,ncb0,ncb1)

	    ## Normal instruction prefetch and pc preincrement
	    self.decodec.OnlyWhen( (5, 6, 7, 8), (self.inst, self.pcincr), (9, ) )

	    ## Additional information for branch case not taken
	    cb0 = ( (self.decodec.syncedto.known[0][0] & self.decodec.case[9]) &
		    (self.bopc.syncedto.known[0][0] & self.bopc.case[1]) )
	    cb1 = ( (self.decodec.syncedto.known[0][1] & self.decodec.case[9]) &
		    (self.bopc.syncedto.known[0][1] & self.bopc.case[1]) )
 
	    ncb0 = ( (self.decodec.syncedto.known[0][0] & self.decodec.case[9]) &
		    (self.bopc.syncedto.known[0][0] & ~self.bopc.case[1]) )
	    ncb1 = ( (self.decodec.syncedto.known[0][1] & self.decodec.case[9]) &
		    (self.bopc.syncedto.known[0][1] & ~self.bopc.case[1]) )

	    self.inst.AddCB(cb0,cb1,ncb0,ncb1)
	    self.pcincr.AddCB(cb0,cb1,ncb0,ncb1)

	    ## Special consideration for pcadj
	    self.decodec.OnlyWhen( (9, ), (self.bop, self.bopc, self.pcadj) )
	    ## Further refinement for pcadj
	    self.pcadj.icb[0] =  self.pcadj.icb[0] & cb0
	    self.pcadj.icb[1] =  self.pcadj.icb[1] & cb1
	    self.pcadj.ncb[0] =  self.pcadj.ncb[0] | ncb0
	    self.pcadj.ncb[1] =  self.pcadj.ncb[1] | ncb1    

	    ## Various jump cases
	    self.decodec.OnlyWhen( (5, 6), (self.pcload, ) )

	    ## Integer stuff
	    self.decodec.OnlyWhen( (0, 1, 2, 3, 7, 8, 9, 10, 11, 12, 13), (self.rdrs, ) )
	    self.hazardc.OnlyWhen2( (0, 2), (self.rdrs, ) )

	    
	    self.decodec.OnlyWhen( (0, 1, 9, 12, 13), (self.rdrt, ) )
	    self.hazardc.OnlyWhen2( (0, 1), (self.rdrt, ) )
	    
	    self.decodec.OnlyWhen( (4, ), (self.rdhl, ) )

	    self.decodec.OnlyWhen( (0, 2 ), (self.intus1, ) )
    
	    self.decodec.OnlyWhen( (1, ), (self.intuc, ) )

	    self.decodec.OnlyWhen( (0, 4, 6, 8, 11), (self.wrrd, ) )
	    self.decodec.OnlyWhen( (2, 10, 11), (self.wrrt, ) )
	    self.decodec.OnlyWhen( (1, 3, ), (self.wrhl, ) )
	    self.decodec.OnlyWhen( (1, ), (self.wrlo, ) )

	    ## Load/Store stuff
	    self.decodec.OnlyWhen( (10, 11), (self.memrd, ) )
	    self.decodec.OnlyWhen( (11, 13), (self.meminc, ) )
	    self.decodec.OnlyWhen( (12, 13), (self.memwr, ) )
	    self.decodec.OnlyWhen( (13, ), (self.rdrf, ) )

	    ## Load/store address computation
	    self.decodec.OnlyWhen( (10, 11, 12, 13), (self.memaddr, ) )


	    ## Bypass out conditions
	    self.decodec.OnlyWhen( (0, 2, 4, 6, 8, 10, 11), (self.bypassout, ))
	    
	def BuildComplete(self,base):

	    print "Complete Build"
	    self.BuildCompositeState(base)
	    Model.BuildCompositeTR(self)
	    Model.BuildCompositecTR(self)
	    #self.BuildOperandLinks()
	    #self.BuildControlStructure()
	    print "Initial CMA TR and cTR:"
	    self.TR.Info()
	    self.cTR.Info()

	    self.TR.SetVarMap()
	    self.BuildVectors((0,0),(0,1)) # Don't flip guards
	    self.BuildTRcTR()
	    print "Dependency/Capacity Refined CMA TR and cTR:"
	    self.TR.Info()
	    self.cTR.Info()

	    outfile = open('dump.dot', 'w')
	    self.DumpDot(outfile)

	    ## Adjust starting state so that long looped pcin and instin do not appear twice
	    ss = [State(-1,self.taskstart[0][0]), State(-1,self.taskstart[0][1])];
	    for i in range(0,len(ss)):
	       ss[i].Smooth(self.pcout.smth)
	       ss[i].Smooth(self.instout.smth)
	       ss[i].Smooth(self.bypassout.smth)

	    self.DefineGroups()

	    rs = self.DualReachTRcTR(ss)
	    self.TR.PruneS(rs)
            ##self.cTR.PruneS(rs)
	    print "Reachable Refined CMA TR and cTR:"     
	    self.TR.Info()
	    self.cTR.Info()

	def BuildShifted(self,base,original):

	    print "Shifted Build"
	    self.BuildCompositeState(base)

	    self.TR = Trans(0,original.TR)
	    self.cTR = Trans(0,original.cTR)

	    shft = ShiftBDD(original.base,original.width+original.base,base)
	    self.TR.Shift(shft)
	    self.cTR.Shift(shft)

	    self.taskstart = [[],]
	    for ss in original.taskstart[0]:
	        ## Copy state
	        temp = State(-1,ss)
		temp.Shift(shft)
		self.taskstart[0].append(temp)

            ## Shift memory care conditions
            i = 0
            self.dual = mgr.One()
            for sm in self.submodels:

                ## Shift dual
                sm.dual = shft.Shift(original.submodels[i].dual)
                self.dual = self.dual & sm.dual
                try:
                        sm.memcare[0][0] = shft.Shift(original.submodels[i].memcare[0][0])
                        sm.memcare[0][1] = shft.Shift(original.submodels[i].memcare[0][1])
                except AttributeError:
                        print "No memorycare for",sm.name
                i = i + 1
            
	    self.forks = [ self.bopc, self.hazardc,  self.blockc, self.decodec ]
	    self.DefineGroups()


	def DefineGroups(self):

            ## Define Groups
	    self.memport_grp = [ self.insts.busy[0], self.inst.busy[0], self.memrd.busy[0], self.memwr.busy[0] ]

	    self.rfrd_grp = [ self.rdrt.busy[0], self.rdrf.busy[0], self.rdrs.busy[0] ]
	    self.rfwr_grp = [ self.wrrd.busy[0], self.wrrt.busy[0] ]

	    self.aludp_grp = [  self.memaddr.busy[0], self.bop.busy[0], self.intus1.busy[0], self.intuc.busy[0], self.meminc.busy[0] ]
	    self.alupc_grp = [  self.pcload.busy[0], self.pcadj.busy[0], self.pcincr.busy[0], self.pcincrs.busy[0] ]

	    self.decode_grp = [ self.decode.busy[0], ]

	    self.arch_grp = [ self.rdhl.busy[0], self.wrhl.busy[0], self.wrlo.busy[0] ]
	  
	    
class Top(Model):

	def __init__(self,parent,iterates):
		
            Model.__init__(self,parent)

            ## Pipe copies
	    self.submodels = []
	    self.iterates = iterates

	    self.submodels.append(RiscPipeStage(0))
	    self.submodels[0].BuildComplete(0)
        
	    for i in range(1,iterates):
		    self.submodels.append(RiscPipeStage(0))
		    base = self.submodels[i-1].width+self.submodels[i-1].base-9 ## 9 for overlap of three operand only machines
		    self.submodels[i].BuildShifted(base,self.submodels[0])

        
	    self.base = 0
	    self.width = self.submodels[iterates-1].base+self.submodels[iterates-1].width

    
	def BuildTR(self):

	    self.forks = []
	    for sm in self.submodels:
		    self.forks = self.forks + sm.forks

	    ## Build resolve labeled state
            vec = DdArray(len(self.forks))
            self.resolve = State(-2)
            for frk in self.forks:
                    self.resolve = self.resolve | frk.resolve



            self.TR = Trans(1) ## Second partition is for long loops
	    self.TR.PreCompose()

	    self.taskstart = [[],]

	    for sm in self.submodels:
	       self.TR.Add(sm.TR)
	       for i in range(0,len(sm.taskstart[0])):
		       try:
			       self.taskstart[0][i] = self.taskstart[0][i] & sm.taskstart[0][i]
		       except IndexError:
			       self.taskstart[0].append(sm.taskstart[0][i])
          
	    self.TR.Compose()

	    ## Discard submodel TR.rel to make room
	    for sm in self.submodels:
	       sm.TR[0].rel= mgr.One()

	    ## Adjust starting state so that long looped pcin and instin do not appear twice
	    for i in range(0,len(self.taskstart[0])):
	       self.taskstart[0][i].Smooth(self.submodels[self.iterates-1].pcout.smth)
	       self.taskstart[0][i].Smooth(self.submodels[self.iterates-1].instout.smth)
	       self.taskstart[0][i].Smooth(self.submodels[self.iterates-1].bypassout.smth)
	       
            ## Top level hold relation
	    self.hold = self.TR.vecps.HoldTR(self.TR.vecns)

	def BuildcTR(self):

	    self.cTR = Trans(1)
	    self.cTR.PreCompose()

	    for sm in self.submodels:
	       self.cTR.Add(sm.cTR)

	    self.cTR.Compose()

	    ## Create second partition for long loops
	    ltr = TransPart()
	
	    
	    ## Adjust smooth vectors
	    ltrsm = [ self.submodels[0].pcin, self.submodels[0].instin, self.submodels[0].bypassin, self.submodels[self.iterates-1].pcout, self.submodels[self.iterates-1].instout, self.submodels[self.iterates-1].bypassout ]

	    for sm in ltrsm:
	        ltr.pssmth[0] = ltr.pssmth[0] & sm.smth.P
		ltr.nssmth[0] = ltr.nssmth[0] & sm.smth.N
		ltr.tagsmth[0] = ltr.tagsmth[0] & sm.smth.T

	    self.cTR[0].pssmth[0] =  self.TR[0].pssmth[0].ExistAbstract(ltr.pssmth[0])
	    self.cTR[0].nssmth[0] =  self.TR[0].nssmth[0].ExistAbstract(ltr.nssmth[0])
	    self.cTR[0].tagsmth[0] =  self.TR[0].tagsmth[0].ExistAbstract(ltr.tagsmth[0])

	    self.cTR.AddTransPart(ltr)

	    ## Add long looping constraints
	    set = []

	    DoubleImplication( set, self.submodels[self.iterates-1].pcout.required[0][0], self.submodels[0].pcin.required[0][1] )
	    DoubleImplication( set, self.submodels[self.iterates-1].pcout.required[0][1], self.submodels[0].pcin.required[0][0] )
	    
	    DoubleImplication( set, self.submodels[self.iterates-1].pcout.known[0][0].P, self.submodels[0].pcin.known[0][1].P )
	    DoubleImplication( set, self.submodels[self.iterates-1].pcout.known[0][1].P, self.submodels[0].pcin.known[0][0].P )

	    DoubleImplication( set, self.submodels[self.iterates-1].instout.required[0][0], self.submodels[0].instin.required[0][1] )
	    DoubleImplication( set, self.submodels[self.iterates-1].instout.required[0][1], self.submodels[0].instin.required[0][0] )
	    
	    DoubleImplication( set, self.submodels[self.iterates-1].instout.known[0][0].P, self.submodels[0].instin.known[0][1].P )
	    DoubleImplication( set, self.submodels[self.iterates-1].instout.known[0][1].P, self.submodels[0].instin.known[0][0].P )

	    DoubleImplication( set, self.submodels[self.iterates-1].bypassout.required[0][0], self.submodels[0].bypassin.required[0][1] )
	    DoubleImplication( set, self.submodels[self.iterates-1].bypassout.required[0][1], self.submodels[0].bypassin.required[0][0] )
	    
	    DoubleImplication( set, self.submodels[self.iterates-1].bypassout.known[0][0].P, self.submodels[0].bypassin.known[0][1].P )
	    DoubleImplication( set, self.submodels[self.iterates-1].bypassout.known[0][1].P, self.submodels[0].bypassin.known[0][0].P )
	    
	    vec = List2Vec(set)
	    self.cTR[1].rel = self.cTR[1].rel & vec.And()

	def BuildGroupsModerate(self):

	    allbounds = mgr.One()

	    memport_grp = []
	    rfrd_grp = []
	    rfwr_grp = []
	    aludp_grp = []
	    alupc_grp = []
	    decode_grp = []
	    arch_grp = []

	    for sm in self.submodels:
		    memport_grp = memport_grp + sm.memport_grp
		    rfrd_grp = rfrd_grp + sm.rfrd_grp
		    rfwr_grp = rfwr_grp + sm.rfwr_grp
		    aludp_grp = aludp_grp + sm.aludp_grp
		    alupc_grp = alupc_grp + sm.alupc_grp
		    decode_grp = decode_grp + sm.decode_grp
		    arch_grp = arch_grp + sm.arch_grp
		     
 
	    
	    ## only pcload is not included in bounding remains unbounded
	    self.memaccess_res = Group(0, 2, memport_grp )
	    allbounds = allbounds & self.memaccess_res.BuildActive()

	    self.rfrd_res = Group(0, 2, rfrd_grp )
	    allbounds = allbounds & self.rfrd_res.BuildActive()
	    
	    self.rfwr_res = Group(0, 1, rfwr_grp )
	    allbounds = allbounds & self.rfwr_res.BuildActive()

	    self.aludp_res = Group(0, 1, aludp_grp )
	    allbounds = allbounds & self.aludp_res.BuildActive()

	    self.alupc_res = Group(0, 1, alupc_grp )
	    allbounds = allbounds & self.alupc_res.BuildActive()

	    self.decode_res = Group(0, 1, decode_grp )
	    allbounds = allbounds & self.decode_res.BuildActive()

	    self.archreg_res = Group(0, 1, arch_grp )
	    allbounds = allbounds & self.archreg_res.BuildActive()

	    self.TR.PruneT(allbounds, (0,1) )

	def BuildGroupsTight(self):

	    allbounds = mgr.One()

	    memport_grp = []
	    rfrd_grp = []
	    rfwr_grp = []
	    aludp_grp = []
	    alupc_grp = []
	    decode_grp = []
	    arch_grp = []

	    for sm in self.submodels:
		    memport_grp = memport_grp + sm.memport_grp
		    rfrd_grp = rfrd_grp + sm.rfrd_grp
		    rfwr_grp = rfwr_grp + sm.rfwr_grp
		    aludp_grp = aludp_grp + sm.aludp_grp
		    alupc_grp = alupc_grp + sm.alupc_grp
		    decode_grp = decode_grp + sm.decode_grp
		    arch_grp = arch_grp + sm.arch_grp

	    ## group read and writes
	    rfrd_grp = rfrd_grp + rfwr_grp
		    
	    ## group alu together
	    aludp_grp = aludp_grp + alupc_grp
 
	    
	    ## only pcload is not included in bounding remains unbounded
	    self.memaccess_res = Group(0, 1, memport_grp )
	    allbounds = allbounds & self.memaccess_res.BuildActive()

	    self.rfrd_res = Group(0, 2, rfrd_grp )
	    allbounds = allbounds & self.rfrd_res.BuildActive()
	    
	    self.rfwr_res = Group(0, 1, rfwr_grp )
	    allbounds = allbounds & self.rfwr_res.BuildActive()

	    self.aludp_res = Group(0, 1, aludp_grp )
	    allbounds = allbounds & self.aludp_res.BuildActive()

	    self.decode_res = Group(0, 1, decode_grp )
	    allbounds = allbounds & self.decode_res.BuildActive()

	    self.archreg_res = Group(0, 1, arch_grp )
	    allbounds = allbounds & self.archreg_res.BuildActive()

	    self.TR.PruneT(allbounds, (0,1) )

    
bm = Top(0,iiterates)

bm.BuildTR()
bm.BuildcTR()
bm.TR.Info()
bm.cTR.Info()

bm.TR.SetVarMap()
bm.BuildVectors((0,0),(0,1)) # Don't flip guards

#order = [39, 40, 41, 42, 43, 44, 45, 46, 47, 4, 3, 48, 49, 50, 7, 6, 2, 5, 8, 9, 10, 11, 87, 88, 89, 90, 91, 92, 60, 61, 62, 57, 58, 59, 32, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 31, 30, 34, 33, 37, 36, 35, 38, 72, 73, 74, 75, 76, 77, 99, 100, 101, 102, 103, 104, 105, 106, 107, 93, 94, 95, 81, 82, 83, 96, 97, 98, 108, 109, 110, 111, 112, 113, 78, 79, 80, 84, 85, 86, 1, 0, 69, 70, 71, 114, 115, 116, 117, 118, 119, 51, 52, 53, 54, 55, 56, 63, 64, 65, 66, 67, 68, 127, 121, 120, 126, 124, 123, 177, 178, 179, 129, 130, 131, 207, 208, 209, 210, 211, 212, 213, 214, 215, 152, 132, 133, 134, 135, 136, 137, 138, 139, 140, 141, 142, 143, 144, 145, 146, 147, 148, 149, 151, 150, 154, 153, 157, 156, 192, 193, 194, 195, 196, 197, 219, 220, 221, 222, 223, 224, 225, 226, 227, 159, 160, 161, 162, 163, 164, 165, 166, 167, 155, 158, 122, 125, 128, 168, 169, 170, 180, 181, 182, 201, 202, 203, 216, 217, 218, 228, 229, 230, 231, 232, 233, 183, 184, 185, 186, 187, 188, 189, 190, 191, 171, 172, 173, 174, 175, 176, 204, 205, 206, 198, 199, 200, 237, 238, 239, 234, 235, 236, 272, 247, 241, 240, 246, 243, 244, 279, 280, 281, 282, 283, 284, 285, 286, 287, 297, 298, 299, 249, 250, 251, 300, 301, 302, 288, 289, 290, 252, 253, 254, 255, 256, 257, 258, 259, 260, 261, 262, 263, 264, 265, 266, 267, 268, 269, 278, 275, 270, 271, 273, 274, 276, 277, 312, 313, 314, 315, 316, 317, 339, 340, 341, 342, 343, 344, 345, 346, 347, 248, 245, 242, 309, 310, 311, 333, 334, 335, 321, 322, 323, 336, 337, 338, 348, 349, 350, 351, 352, 353, 291, 292, 293, 294, 295, 296, 303, 304, 305, 306, 307, 308, 318, 319, 320, 324, 325, 326, 354, 355, 356, 357, 358, 359, 327, 328, 329, 330, 331, 332, 368, 367, 365, 362, 361, 360, 364, 363, 366]

#bm.ImportOrder(order)

bm.Reorder()
bm.TR.Info()
bm.cTR.Info()

rs = bm.DualReachTRcTR(bm.taskstart[0])
	
bm.TR.PruneS(rs,(0,1))
bm.cTR.Collapse()
bm.TR.Info()
bm.cTR.Info()

if iconfig=="t":
	bm.BuildGroupsTight()
	bm.TR.Info()

	bm.Reorder()
	bm.TR.Info()
	bm.cTR.Info()
elif iconfig=="m":
	bm.BuildGroupsModerate()
	bm.TR.Info()

	bm.Reorder()
	bm.TR.Info()
	bm.cTR.Info()
	
bm.pin = bm.submodels[1].decode
result = bm.CSchedule()

print os.times()

execfile(mix)

print os.times()


















